// file: main.C

// Program to estimate genotypes and allele frequencies for a
// population where genotype is based on evidence from NGS reads.

// Time-stamp: <Sunday, 01 May 2011, 16:41 MDT -- zgompert>

#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include <time.h>
#include "alleleEst.h"

// compile on OSX

// g++ -O2 -Wall -o alleleEst main.C func.C -lgsl -lm -framework Accelerate

// compile on seismic
// g ++ -O2 -I/usr/local/gsl-1.14/include -I/usr/local/atlas-3.9.28/include  -L/usr/local/gsl-1.14/lib -L/usr/local/atlas-3.9.28/lib -Wall -o bgc bgc_main.C bgc_func_readdata.C bgc_func_initialize.C bgc_func_mcmc.C bgc_func_write.C bgc_func_linkage.C bgc_func_ngs.C mvrandist.c  -lgsl -lcblas -latlas -lm 

using namespace std;

gsl_rng * r;  /* global state variable for random number generator */

/* ----------------- */
/* beginning of main */
/* ----------------- */

int main(int argc, char *argv[]) {

  time_t start = time(NULL);
  time_t end;
  int rng_seed = 0, ch = 0, i = 0, j = 0, step = 0, mcmcL= 1;
  int thin = 1;
  int ctr = 0;

  string readsfile = "undefined";
  string afreqof = "of_afreq.txt";
  string gprobof = "of_gprob.txt";
  paramcont params;
  datacont data;
  datadimcont datadim;
  initcont initvals;

  FILE * fp; // use for outfiles

  initvals.alpha = 1;

  // get command line arguments
  if (argc < 2) {
    usage(argv[0]);
  }
  
  while ((ch = getopt(argc, argv, "l:g:t:a:o:")) != -1){
    switch(ch){
    case 'l':
      mcmcL = atoi(optarg);
      break;
    case 'g':
      readsfile = optarg;
      break;
    case 't':
      thin = atoi(optarg);
      break;
    case 'a':
      afreqof = optarg;
      break;
    case 'o':
      gprobof = optarg;
      break;
    case '?':
    default:
      usage(argv[0]);
    }
  }

  // set up gsl random number generation 
  gsl_rng_env_setup();
  r = gsl_rng_alloc (gsl_rng_default);
  srand(time(NULL));
  rng_seed = rand();
  gsl_rng_set(r, rng_seed); /* seed gsl_rng with output of rand, which
			       was seeded with result of time(NULL) */
  
  cerr << "Reading input files" << endl;

  // determine number of loci and number of alleles per locus
  datadim.nloci = getNloci(readsfile);
  datadim.nallele = gsl_vector_int_calloc(datadim.nloci);
  datadim.maxNallele = getNallele(readsfile, datadim.nallele);

  params.alleleparams = gsl_vector_calloc(datadim.maxNallele);
  params.alleleresult = gsl_vector_calloc(datadim.maxNallele);
  params.allelecount = gsl_vector_uint_calloc(datadim.maxNallele);

  // determine number of individuals per locus
  datadim.nind = getNind(readsfile);
  cerr << "Number of loci: " << datadim.nloci << endl;
  cerr << "Number of individuals: " << datadim.nind << endl;

  //allocate memory to store genetics parameters (indicator variable)
  params.genetics = new mat_container_int * [datadim.nloci];
  for(i=0; i<datadim.nloci; i++){
    params.genetics[i] = new mat_container_int[datadim.nind];
    for(j=0; j<datadim.nind ; j++){
      params.genetics[i][j].mat = gsl_matrix_int_calloc(gsl_vector_int_get(datadim.nallele,i), 
						 gsl_vector_int_get(datadim.nallele,i));
    }
  }
  // allocate memory for raw read data for loci and inds
  data.reads = new mat_container_int[datadim.nloci];
  for(i=0; i<datadim.nloci; i++){
    data.reads[i].mat = gsl_matrix_int_calloc(datadim.nind, gsl_vector_int_get(datadim.nallele,i));
  }
  data.nonzeroalleles = gsl_matrix_int_calloc(datadim.nloci, datadim.nind);

  // allocate memory for paramters
  params.popallelefreq = gsl_matrix_calloc(datadim.nloci, datadim.maxNallele);

  //allocate memory to store probability of each genotypes (MCMC samples)
  params.prgenetics = new mat_container * [datadim.nloci];
  for(i=0; i<datadim.nloci; i++){
    params.prgenetics[i] = new mat_container [datadim.nind];
    for(j=0; j<datadim.nind ; j++){
      params.prgenetics[i][j].mat = gsl_matrix_calloc(gsl_vector_int_get(datadim.nallele,i), 
						 gsl_vector_int_get(datadim.nallele,i));
    }
  }

  // read in data
  getreads(readsfile, &data);
  cerr << "Initializing chain" << endl;
  initparams(&datadim, &params, &initvals, &data);
  
  // open allele freq outfile
  fp = fopen(afreqof.c_str(), "w");

  // MCMC
  cerr << "Running chain ";
  for(step=0; step<mcmcL; step++){
    mcmcUpdate(&datadim, &params, &data);

    if((step % thin) == 0){
      // print mcmc output
      writeSamples(&datadim, &params, fp);
      // running sum of genotype probs.
      calcpGenotype(&datadim, &params);
      ctr++;
    }
    if((step % 1000) == 0){
      cerr << "*";
    }
  }
  cerr << endl << "Writing final results" << endl;
  fclose(fp); // done writing allele freq.
  // write genotypes
  fp = fopen(gprobof.c_str(), "w");
  writeGen(&datadim, &params, fp, ctr);
  fclose(fp);

  // free dynamic memory ---------------------
  gsl_vector_int_free(datadim.nallele);
  for(i=0; i<datadim.nloci; i++){
    for(j=0; j<datadim.nind ; j++){
      gsl_matrix_int_free(params.genetics[i][j].mat);
      gsl_matrix_free(params.prgenetics[i][j].mat);
    }
    delete params.genetics[i];
    delete params.prgenetics[i];
  }
  delete params.genetics;
  delete params.prgenetics;

  for(i=0; i<datadim.nloci; i++){
    gsl_matrix_int_free(data.reads[i].mat);
  }
  delete data.reads;
  delete params.popallelefreq;

  // -----------------------------------------

  // prints run time
  end = time(NULL);
  cerr << "Runtime: " << (end-start)/3600 << " hr " << (end-start)%3600/60 << " min ";
  cerr << (end-start)%60 << " sec" << endl;
  return 0;
}
